
// Author:  David Martin  <dmartin@eecs.berkeley.edu>

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;
import java.net.*;

import java.util.List;

// This class is 100% UI.
// It delegates all operations on data to other classes.
public class MainWindow
    extends JFrame
    implements ActionListener
{
    // If this flag is true, the segments outlines are drawn.
    static boolean showSegments = true;

    // If this flag is true, the image is drawn.
    static boolean showImage = true;

    // Remember the places where the user is saving and opening files.
    String chooserSegDir = null;
    String chooserImageDir = null;
    String chooserURL = "http://";

    // Magnification.
    static int border = 10;
    static boolean fitWindow = true;
    static double magnification = 1.0;

    // UI Elements.
    JTextField magTextField;
    JCheckBox fitWindowCheckBox;
    SegmentationPanel segPanel;
    JMenuItem fileMenuNewImage;
    JMenuItem fileMenuChangeUserID;
    JMenuItem fileMenuOpenImageFile;
    JMenuItem fileMenuOpenImageURL;
    JMenuItem fileMenuOpenImageByID;
    JMenuItem fileMenuSaveSeg;
    JMenuItem fileMenuSaveAsSeg;
    JMenuItem fileMenuOpenSegFile;
    JMenuItem fileMenuOpenSegURL;
    JMenuItem fileMenuSubmitSeg;
    JMenuItem fileMenuOpenSubmittedSeg;
    JMenuItem fileMenuOpenSubmittedSegByID;
    JMenuItem fileMenuAbortSeg;
    JMenuItem fileMenuQuit;
    JMenuItem editMenuUndo, editMenuRedo;
    JMenuItem viewMenuRefresh;
    JCheckBoxMenuItem viewMenuGray, viewMenuFlip, viewMenuInvert;
    JCheckBoxMenuItem viewMenuSegments, viewMenuImage, viewMenuFitWindow;
    JMenuItem viewMenuMag1x, viewMenuMag2x, viewMenuMag3x, viewMenuMag4x;
    JMenuItem helpMenuFiles, helpMenuEdit;
    JLabel imageNameVal, segNameVal, userNameVal, totalImageCountVal, sessionImageCountVal, segCountVal;
    SplitterWindow splitterWindow;
    JScrollPane scrollPane;

    public MainWindow (String title) {
	super (title);

	JMenu fileMenu = new JMenu ("File");
	fileMenuChangeUserID = new JMenuItem ("Change User ID");
	fileMenuNewImage = new JMenuItem ("Get New Image");
	fileMenuOpenImageByID = new JMenuItem ("Get Image by ID...");
	fileMenuOpenImageFile = new JMenuItem ("Open Image File...");
	fileMenuOpenImageURL = new JMenuItem ("Open Image URL...");
	fileMenuSaveSeg = new JMenuItem ("Save Local Copy");
	fileMenuSaveAsSeg = new JMenuItem ("Save Local Copy As...");
	if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
	    fileMenuOpenSegFile = new JMenuItem ("Open Local Copy...");
	} else {
	    fileMenuOpenSegFile = new JMenuItem ("Open Segmentation File...");
	}
	fileMenuOpenSegURL = new JMenuItem ("Open Segmentation URL...");
	fileMenuAbortSeg = new JMenuItem ("Abort Segmentation");
	fileMenuSubmitSeg = new JMenuItem ("Submit Segmentation");
	fileMenuOpenSubmittedSeg = new JMenuItem ("Open Submitted Segmentation...");
	fileMenuOpenSubmittedSegByID = new JMenuItem ("Open Submitted Segmentation by ID...");
	fileMenuQuit = new JMenuItem ("Quit");
	fileMenuNewImage.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_N, ActionEvent.CTRL_MASK));
	fileMenuOpenSegFile.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_O, ActionEvent.CTRL_MASK));
	fileMenuSaveSeg.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_S, ActionEvent.CTRL_MASK));
	fileMenuSaveAsSeg.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_A, ActionEvent.CTRL_MASK));
	fileMenuAbortSeg.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_B, ActionEvent.CTRL_MASK));
	fileMenuSubmitSeg.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_T, ActionEvent.CTRL_MASK));
	fileMenuQuit.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_Q, ActionEvent.CTRL_MASK));

	if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
	    if (SegmentApp.superUser) {
		fileMenu.add (fileMenuChangeUserID);
		fileMenu.add (fileMenuOpenImageByID);
		fileMenu.add (fileMenuOpenSegURL);
		fileMenu.addSeparator ();
	    }
	    fileMenu.add (fileMenuNewImage);
	    fileMenu.add (fileMenuOpenSubmittedSeg);
	    fileMenu.add (fileMenuOpenSubmittedSegByID);
	    fileMenu.addSeparator ();
	    fileMenu.add (fileMenuSaveSeg);
	    fileMenu.add (fileMenuSaveAsSeg);
	    fileMenu.add (fileMenuOpenSegFile);
	    fileMenu.addSeparator ();
	    fileMenu.add (fileMenuAbortSeg);
	    fileMenu.add (fileMenuSubmitSeg);
	    fileMenu.addSeparator ();
	    fileMenu.add (fileMenuQuit);
	} else {
	    fileMenu.add (fileMenuOpenImageFile);
	    fileMenu.add (fileMenuOpenImageURL);
	    fileMenu.addSeparator ();
	    fileMenu.add (fileMenuOpenSegFile);
	    fileMenu.add (fileMenuOpenSegURL);
	    fileMenu.addSeparator ();
	    fileMenu.add (fileMenuSaveSeg);
	    fileMenu.add (fileMenuSaveAsSeg);
	    fileMenu.addSeparator ();
	    fileMenu.add (fileMenuQuit);
	}

	fileMenuChangeUserID.addActionListener (this);
	fileMenuNewImage.addActionListener (this);
	fileMenuOpenImageByID.addActionListener (this);
	fileMenuOpenImageFile.addActionListener (this);
	fileMenuOpenImageURL.addActionListener (this);
	fileMenuSaveSeg.addActionListener (this);
	fileMenuSaveAsSeg.addActionListener (this);
	fileMenuOpenSegFile.addActionListener (this);
	fileMenuOpenSegURL.addActionListener (this);
	fileMenuAbortSeg.addActionListener (this);
	fileMenuSubmitSeg.addActionListener (this);
	fileMenuOpenSubmittedSeg.addActionListener (this);
	fileMenuOpenSubmittedSegByID.addActionListener (this);
	fileMenuQuit.addActionListener (this);

	JMenu editMenu = new JMenu ("Edit");
	editMenuUndo = new JMenuItem ("Undo");
	editMenuRedo = new JMenuItem ("Redo");
	editMenuUndo.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_Z, ActionEvent.CTRL_MASK));
	editMenuRedo.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_R, ActionEvent.CTRL_MASK));
	editMenu.add (editMenuUndo);
	editMenu.add (editMenuRedo);
	editMenuUndo.addActionListener (this);
	editMenuRedo.addActionListener (this);

	JMenu viewMenu = new JMenu ("View");
	viewMenuRefresh = new JMenuItem ("Refresh Window");
	viewMenuGray = new JCheckBoxMenuItem ("Gray Image");
	viewMenuFlip = new JCheckBoxMenuItem ("Flip Image");
	viewMenuInvert = new JCheckBoxMenuItem ("Invert Image");
	viewMenuSegments = new JCheckBoxMenuItem ("Show Segments");
	viewMenuImage = new JCheckBoxMenuItem ("Show Image");
	viewMenuFitWindow = new JCheckBoxMenuItem ("Fit Window");
	viewMenuMag1x = new JMenuItem ("Magnify 1x");
	viewMenuMag2x = new JMenuItem ("Magnify 2x");
	viewMenuMag3x = new JMenuItem ("Magnify 3x");
	viewMenuMag4x = new JMenuItem ("Magnify 4x");
	viewMenuGray.setSelected (SegmentApp.grayImage);
	viewMenuFlip.setSelected (SegmentApp.flipImage);
	viewMenuInvert.setSelected (SegmentApp.invertImage);
	viewMenuSegments.setSelected (showSegments);
	viewMenuImage.setSelected (showImage);
	viewMenuFitWindow.setSelected (fitWindow);
	viewMenuRefresh.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_L, ActionEvent.CTRL_MASK));
	viewMenuSegments.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_W, ActionEvent.CTRL_MASK));
	viewMenuImage.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_E, ActionEvent.CTRL_MASK));
	viewMenuFitWindow.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_F, ActionEvent.CTRL_MASK));
	viewMenuMag1x.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_1, ActionEvent.CTRL_MASK));
	viewMenuMag2x.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_2, ActionEvent.CTRL_MASK));
	viewMenuMag3x.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_3, ActionEvent.CTRL_MASK));
	viewMenuMag4x.setAccelerator (KeyStroke.getKeyStroke (
	    KeyEvent.VK_4, ActionEvent.CTRL_MASK));
	if (SegmentApp.superUser) {
	    viewMenu.add (viewMenuGray);
	    viewMenu.add (viewMenuFlip);
	    viewMenu.add (viewMenuInvert);
	    viewMenu.addSeparator ();
	}
	viewMenu.add (viewMenuRefresh);
	viewMenu.add (viewMenuSegments);
	viewMenu.add (viewMenuImage);
	viewMenu.addSeparator ();
	viewMenu.add (viewMenuMag1x);
	viewMenu.add (viewMenuMag2x);
	viewMenu.add (viewMenuMag3x);
	viewMenu.add (viewMenuMag4x);
	viewMenu.add (viewMenuFitWindow);
	viewMenuRefresh.addActionListener (this);
	viewMenuGray.addActionListener (this);
	viewMenuFlip.addActionListener (this);
	viewMenuInvert.addActionListener (this);
	viewMenuSegments.addActionListener (this);
	viewMenuImage.addActionListener (this);
	viewMenuMag1x.addActionListener (this);
	viewMenuMag2x.addActionListener (this);
	viewMenuMag3x.addActionListener (this);
	viewMenuMag4x.addActionListener (this);
	viewMenuFitWindow.addActionListener (this);

	JMenu helpMenu = new JMenu ("Help");
	helpMenuFiles = new JMenuItem ("Files");
	helpMenuEdit = new JMenuItem ("Editing");
	helpMenu.add (helpMenuFiles);
	helpMenu.add (helpMenuEdit);
	helpMenuFiles.addActionListener (this);
	helpMenuEdit.addActionListener (this);

	JMenuBar mbar = new JMenuBar ();
	mbar.add (fileMenu);
	mbar.add (editMenu);
	mbar.add (viewMenu);
	mbar.add (Box.createHorizontalGlue ());
	mbar.add (helpMenu);
	setJMenuBar (mbar);

	JLabel userNameLab = new JLabel ("User :");
	JLabel imageNameLab = new JLabel ("Current Image :");
	JLabel segNameLab = new JLabel ("Current Seg File:");
	JLabel totalImageCountLab = new JLabel ("Total Images Completed :");
	JLabel sessionImageCountLab = new JLabel ("Images This Session :");
	JLabel segCountLab = new JLabel ("Segments :");
	JLabel magLabel = new JLabel ("Magnification :");
	userNameVal = new JLabel ("");
	imageNameVal = new JLabel ("");
	segNameVal = new JLabel ("");
	totalImageCountVal = new JLabel ("");
	sessionImageCountVal = new JLabel ("");
	segCountVal = new JLabel ("");
	magTextField = new JTextField (5);
	fitWindowCheckBox = new JCheckBox ("Fit Window", fitWindow);
	userNameLab.setForeground (Color.black);
	imageNameLab.setForeground (Color.black);
	segNameLab.setForeground (Color.black);
	totalImageCountLab.setForeground (Color.black);
	sessionImageCountLab.setForeground (Color.black);
	segCountLab.setForeground (Color.black);
	magLabel.setForeground (Color.black);
	magTextField.addActionListener (this);
	fitWindowCheckBox.addActionListener (this);

	Box box1 = Box.createHorizontalBox ();
	Box box2 = Box.createHorizontalBox ();
	
	if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
	    box1.add (Box.createHorizontalStrut (10));
	    box1.add (userNameLab);
	    box1.add (Box.createHorizontalStrut (5));
	    box1.add (userNameVal);
	    box1.add (Box.createHorizontalStrut (5));
	    box1.add (Box.createHorizontalGlue ());
	    box1.add (totalImageCountLab);
	    box1.add (Box.createHorizontalStrut (5));
	    box1.add (totalImageCountVal);
	    box1.add (Box.createHorizontalStrut (10));

	    box2.add (Box.createHorizontalStrut (10));
	    box2.add (imageNameLab);
	    box2.add (Box.createHorizontalStrut (5));
	    box2.add (imageNameVal);
	    box2.add (Box.createHorizontalStrut (10));
	    box2.add (Box.createHorizontalGlue ());
	    box2.add (segCountLab);
	    box2.add (Box.createHorizontalStrut (5));
	    box2.add (segCountVal);
	    box2.add (Box.createHorizontalStrut (10));
	    box2.add (Box.createHorizontalGlue ());
	    box2.add (sessionImageCountLab);
	    box2.add (Box.createHorizontalStrut (5));
	    box2.add (sessionImageCountVal);
	    box2.add (Box.createHorizontalStrut (10));
	} else {
	    box1.add (Box.createHorizontalStrut (10));
	    box1.add (imageNameLab);
	    box1.add (Box.createHorizontalStrut (5));
	    box1.add (imageNameVal);
	    box1.add (Box.createHorizontalStrut (10));
	    box1.add (Box.createHorizontalGlue ());
	    box1.add (segCountLab);
	    box1.add (Box.createHorizontalStrut (5));
	    box1.add (segCountVal);
	    box1.add (Box.createHorizontalStrut (10));

	    box2.add (Box.createHorizontalStrut (10));
	    box2.add (segNameLab);
	    box2.add (Box.createHorizontalStrut (5));
	    box2.add (segNameVal);
	    box2.add (Box.createHorizontalGlue ());
	}

	JPanel magTextBox = new JPanel ();
	magTextBox.setLayout (new FlowLayout (FlowLayout.LEFT));
	magTextBox.add (magTextField);

	Box magBox = Box.createHorizontalBox ();
	magBox.add (Box.createHorizontalStrut (10));
	magBox.add (magLabel);
	magBox.add (Box.createHorizontalStrut (5));
	magBox.add (magTextBox);
	magBox.add (Box.createHorizontalGlue ());
	magBox.add (Box.createHorizontalStrut (5));
	magBox.add (fitWindowCheckBox);
	magBox.add (Box.createHorizontalStrut (5));
	JPanel magPanel = new JPanel ();
	magPanel.setLayout (new BorderLayout ());
	magPanel.add (magBox, BorderLayout.NORTH);
	magPanel.setBorder (new BevelBorder (BevelBorder.RAISED));

	Box topBox1 = Box.createVerticalBox  ();
	topBox1.add (Box.createVerticalStrut (5));
	topBox1.add (box1);
	topBox1.add (Box.createVerticalStrut (5));
	topBox1.add (box2);
	topBox1.add (Box.createVerticalStrut (5));

	JPanel topPanel1 = new JPanel ();
	topPanel1.setLayout (new BorderLayout ());
	topPanel1.add (topBox1, BorderLayout.NORTH);
	topPanel1.setBorder (new BevelBorder (BevelBorder.RAISED));

	Box topPanel = Box.createVerticalBox  ();
	topPanel.add (topPanel1);
	topPanel.add (magPanel);

	segPanel = new SegmentationPanel ();
	scrollPane = new JScrollPane (
	    segPanel, 
	    JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
	    JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
	scrollPane.setBorder (new BevelBorder (BevelBorder.RAISED));

	JPanel mainPanel = new JPanel ();
	mainPanel.setLayout (new BorderLayout ());
	mainPanel.add (topPanel, BorderLayout.NORTH);
	mainPanel.add (scrollPane, BorderLayout.CENTER);

	setContentPane (mainPanel);

	// Make sure we can prevent the user from blowing away the
	// window with unsaved changes.
	setDefaultCloseOperation (JFrame.DO_NOTHING_ON_CLOSE);
	addWindowListener (new WindowAdapter () {
	    public void windowClosing (WindowEvent e) {
		SegmentApp.quit ();
	    };
	});

	splitterWindow = new SplitterWindow (this, "Pixel Exchange");

	pack ();	// Must call pack before setSize.
	setSize (500, 800);
	setLocation (0, 0);

	scrollPane.getViewport().addComponentListener (
	    new ComponentAdapter () {
		    public void componentResized (ComponentEvent e) {
			updateImageDisplay ();
		    }
		});
    }

    public void newImage () {
	splitterWindow.newImage ();
    }
    
    // Preconditions: One segment selected.
    // Actions: Open the splitter window to split the selected segment.
    void splitSegment () {
	Util.assert (SegmentApp.imageLoaded());
	Util.assert (SegmentApp.selection.countSelected() == 1);

	int[] sel = SegmentApp.selection.getSelectedIDs ();
	int newSegID = SegmentApp.segmentation.newSegmentID ();

	SegmentApp.selection.toggleSelection (newSegID);
	splitterWindow.setSegments (sel[0], newSegID);
	splitterWindow.show ();
    }

    // Preconditions: Two segments are selected.
    // Actions: Open the splitter window to split the two selected segments.
    void editSegments () {
	Util.assert (SegmentApp.imageLoaded());
	Util.assert (SegmentApp.selection.countSelected() == 2);

	int[] sel = SegmentApp.selection.getSelectedIDs ();
	splitterWindow.setSegments (sel[0], sel[1]);
	splitterWindow.show ();
    }

    // Save current segmentation to currently opened segmentation file.
    // Return true if the file was successfully saved.
    void saveSegmentation () {
	Util.assert (SegmentApp.imageLoaded());
	try {
	    String fileName;
	    if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
		fileName = SegmentApp.userSegDir
			   + "/" + SegmentApp.imageID + ".seg";
	    } else {
		if (SegmentApp.segFileName == null) {
		    saveAsSegmentation ();
		    return;
		}
		fileName = SegmentApp.segFileName;
	    }
	    SegmentApp.saveSegmentation (new File (fileName));
	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error saving segmentation: " + e.getMessage(),
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    // Save current segmentation to new file.
    // Return true if the file was successfully saved.
    void saveAsSegmentation () {
	Util.assert (SegmentApp.imageLoaded());
	try {
	    if (chooserSegDir == null) {
		try {
		    chooserSegDir = SegmentApp.userSegDir;
		} catch (Throwable e) {
		    chooserSegDir = null;
		}
	    }
	    JFileChooser chooser = new JFileChooser ();
	    chooser.setFileSelectionMode (JFileChooser.FILES_ONLY);
	    chooser.addChoosableFileFilter (new SegFileFilter ());
	    if (chooserSegDir != null) {
		chooser.setCurrentDirectory (new File (chooserSegDir));
	    }
	    int rval = chooser.showSaveDialog (this);
	    if (rval == JFileChooser.APPROVE_OPTION) {
		// User selected a file.
		File file = chooser.getSelectedFile();
		chooserSegDir = file.getAbsolutePath ();
		// Warn about overwriting an existing file.
		if (file.exists()) {
		    int answer = JOptionPane.showConfirmDialog (
			this, Util.lineBreak (
			    "The file " + file + " already exists.\n" +
			    "Do you want to overwrite it?",
			    SegmentApp.infoWidth),
			"Question", JOptionPane.YES_NO_OPTION);
		    if (answer == JOptionPane.NO_OPTION) {
			return;
		    }
		}
		SegmentApp.saveSegmentation (file);
		SegmentApp.segFileName = file.toString();
	    }
	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error saving segmentation: " + e.getMessage(), 
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    static class SegFileFilter extends javax.swing.filechooser.FileFilter {
	public boolean accept (File f) {
	    if (f.isDirectory()) { return true; }
	    String name = f.getName ();
	    if (name.endsWith (".seg")) { return true; }
	    if (name.endsWith (".SEG")) { return true; }
	    return false;
	}
	public String getDescription () {
	    return ("Segmentation Files (*.seg)");
	}
    }

    static class ImageFileFilter extends javax.swing.filechooser.FileFilter {
	public boolean accept (File f) {
	    if (f.isDirectory()) { return true; }
	    String name = f.getName ();
	    if (name.endsWith (".jpg")) { return true; }
	    if (name.endsWith (".jpeg")) { return true; }
	    if (name.endsWith (".jpe")) { return true; }
	    if (name.endsWith (".jpa")) { return true; }
	    if (name.endsWith (".gif")) { return true; }
	    return false;
	}
	public String getDescription () {
	    return ("Image Files (*.jpg *.jpeg *.jpe *.jpa *.gif)");
	}
    }

    // Open a new segmentation and associated image.
    void openSegmentationFile () {
	try {
	    if (SegmentApp.unsavedChanges) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"There are unsaved changes which will be lost " +
			"if this operation continues! " +
			"Are you sure you want to open a new segmentation before " +
			"first saving the current segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } else if (SegmentApp.mode == SegmentApp.CLIENT_MODE && SegmentApp.imageLoaded()) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"Are you sure you want to open a new segmentation before " +
			"first submitting or aborting the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } 
	    if (chooserSegDir == null) {
		try {
		    chooserSegDir = SegmentApp.userSegDir;
		} catch (Throwable e) {
		    chooserSegDir = null;
		}
	    }
	    JFileChooser chooser = new JFileChooser ();
	    chooser.setFileSelectionMode (JFileChooser.FILES_ONLY);
	    chooser.addChoosableFileFilter (new SegFileFilter ());
	    if (chooserSegDir != null) {
		chooser.setCurrentDirectory (new File (chooserSegDir));
	    }
	    int rval = chooser.showOpenDialog (this);
	    if (rval == JFileChooser.APPROVE_OPTION) {
		// User selected a file.
		File file = chooser.getSelectedFile();
		chooserSegDir = file.getAbsolutePath ();
		Segmentation seg = Segmentation.openSegmentationFile (file);
		if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
		    int imageID = seg.getImageID ();
		    Image image = SegmentApp.loadImageByID (imageID);
		    SegmentApp.installNewImageAndSegmentation (
			image, seg, null, null);
		} else {
		    Util.assert (SegmentApp.imageLoaded());
		    SegmentApp.installNewSegmentation (
			seg, file.toString());
		}
	    }
	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error opening segmentation: " + e.getMessage(), 
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }
    
    // Open a new segmentation and associated image.
    void openSegmentationURL () {
	try {
	    if (SegmentApp.unsavedChanges) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"There are unsaved changes which will be lost " +
			"if this operation continues! " +
			"Are you sure you want to open a new segmentation before " +
			"first saving the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } else if (SegmentApp.mode == SegmentApp.CLIENT_MODE && SegmentApp.imageLoaded()) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"Are you sure you want to open a new segmentation before " +
			"first submitting or aborting the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } 

	    String s = (String) JOptionPane.showInputDialog (
		this, "Enter the segmentation URL.",
		"Open Segmentation URL", JOptionPane.INFORMATION_MESSAGE,
		null, null, chooserURL);
	    if (s == null) {
		return;	// User pressed cancel button in InputDialog.
	    }
	    chooserURL = s;
	    URL url = new URL (s);
	    Segmentation seg = Segmentation.openSegmentationURL (url);
	    if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
		int imageID = seg.getImageID ();
		Image image = SegmentApp.loadImageByID (imageID);
		SegmentApp.installNewImageAndSegmentation (
		    image, seg, null, null);
	    } else {
		Util.assert (SegmentApp.imageLoaded());
		SegmentApp.installNewSegmentation (seg, null);
	    }

	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error opening segmentation: " + e.getMessage(), 
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }
    
    // Open a previously submitted segmentation file and associated image.
    void openSubmittedSegmentation () {
	Util.assert (SegmentApp.mode == SegmentApp.CLIENT_MODE);
	try {
	    if (SegmentApp.unsavedChanges) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"There are unsaved changes which will be lost " +
			"if this operation continues! " +
			"Are you sure you want to open a new segmentation before " +
			"first saving the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } else if (SegmentApp.mode == SegmentApp.CLIENT_MODE && SegmentApp.imageLoaded()) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"Are you sure you want to open a new segmentation before " +
			"first submitting or aborting the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } 
	
	    // Get list of image IDs from server.
	    List imageIDList = SegmentApp.getSubmittedSegList ();
	    if (imageIDList == null) { 
		JOptionPane.showMessageDialog (
		    this, Util.lineBreak (
			"Error getting submitted image list from server.", 
			SegmentApp.infoWidth), 
		    "Error", JOptionPane.ERROR_MESSAGE);
		return;
	    }

	    if (imageIDList.size() == 0) {
		JOptionPane.showMessageDialog (
		    this, Util.lineBreak (
			"You have not submitted any segmentations yet!", 
			SegmentApp.infoWidth), 
		    "Message", JOptionPane.INFORMATION_MESSAGE);
		return;
	    }

	    // Have user pick one of the IDs.
	    ImageChooser imageChooser = new ImageChooser (
		this, "Chooser Image ID", imageIDList);
	    imageChooser.show ();
	    int imageID = imageChooser.getImageID ();
	    if (imageID < 0) {
		// User pressed cancel.
		return;
	    }

	    // Load the segmentation for the ID the user chose.
	    Segmentation seg = SegmentApp.openSegmentationByID (imageID);
	    Image image = SegmentApp.loadImageByID (imageID);
	    SegmentApp.installNewImageAndSegmentation (
		image, seg, null, null);

	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error opening segmentation: " + e.getMessage(), 
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }
    
    // Open a previously submitted segmentation file and associated image.
    void openSubmittedSegmentationByID () {
	Util.assert (SegmentApp.mode == SegmentApp.CLIENT_MODE);
	try {
	    if (SegmentApp.unsavedChanges) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"There are unsaved changes which will be lost " +
			"if this operation continues! " +
			"Are you sure you want to open a new segmentation before " +
			"first saving the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } else if (SegmentApp.mode == SegmentApp.CLIENT_MODE && SegmentApp.imageLoaded()) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"Are you sure you want to open a new segmentation before " +
			"first submitting or aborting the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } 
	
	    // Get the image ID from the user.
	    String s = (String) JOptionPane.showInputDialog (
		this, "Enter the image ID.",
		"Open Submitted Segmentation By ID", JOptionPane.INFORMATION_MESSAGE,
		null, null, "");
	    if (s == null) {
		// User pressed cancel button in InputDialog.
		return;
	    }
	    int imageID = Integer.decode(s).intValue();
	    if (imageID < 0) {
		throw new Exception ("Invalid image ID: " + imageID + ".");
	    }

	    // Load the segmentation for the ID the user chose.
	    Segmentation seg = SegmentApp.openSegmentationByID (imageID);
	    Image image = SegmentApp.loadImageByID (imageID);
	    SegmentApp.installNewImageAndSegmentation (
		image, seg, null, null);

	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error opening segmentation: " + e.getMessage(), 
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }
    
    void getNewImage () {
	Util.assert (SegmentApp.mode == SegmentApp.CLIENT_MODE);
	try {
	    if (SegmentApp.unsavedChanges) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"There are unsaved changes which will be lost " +
			"if this operation continues! " +
			"Are you sure you want to acquire a new image before " +
			"first saving the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } else if (SegmentApp.mode == SegmentApp.CLIENT_MODE && SegmentApp.imageLoaded()) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"Are you sure you want to acquire a new image before " +
			"first submitting or aborting the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } 
	    SegmentApp.ImageRequest ir = SegmentApp.getNewImageID ();
	    if (ir.allDone) {
		throw new Exception ("You have segmented all available images!");
	    }
	    Image image = SegmentApp.loadImageByID (ir.imageID);
	    SegmentApp.installNewImage (ir.imageID, 
					ir.gray, ir.invert, ir.flipflop,
					image, null);
	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error opening new image: " + e.getMessage(),
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    void openImageFile () {
	Util.assert (SegmentApp.mode == SegmentApp.HERMIT_MODE);
	try {
	    if (SegmentApp.unsavedChanges) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"There are unsaved changes which will be lost " +
			"if this operation continues! " +
			"Are you sure you want to open a new image before " +
			"first saving the current segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    }
	    if (chooserImageDir == null) {
		try {
		    chooserImageDir = System.getProperty ("user.home");
		} catch (Throwable e) {
		    chooserImageDir = null;
		}
	    }
	    JFileChooser chooser = new JFileChooser ();
	    chooser.setFileSelectionMode (JFileChooser.FILES_ONLY);
	    chooser.addChoosableFileFilter (new ImageFileFilter ());
	    if (chooserImageDir != null) {
		chooser.setCurrentDirectory (new File (chooserImageDir));
	    }
	    int rval = chooser.showOpenDialog (this);
	    if (rval == JFileChooser.APPROVE_OPTION) {
		// User selected a file.
		File file = chooser.getSelectedFile();
		chooserImageDir = file.getAbsolutePath ();
		Image image = Util.loadImage (file);
		SegmentApp.installNewImage (
		    SegmentApp.INVALID_IMAGE_ID, 
		    SegmentApp.grayImage, 
		    SegmentApp.invertImage,
		    SegmentApp.flipImage,
		    image, file.getName());
	    }

	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error opening image: " + e.getMessage(),
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    void openImageURL () {
	Util.assert (SegmentApp.mode == SegmentApp.HERMIT_MODE);
	try {
	    if (SegmentApp.unsavedChanges) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"There are unsaved changes which will be lost " +
			"if this operation continues! " +
			"Are you sure you want to open a new image before " +
			"first saving the current segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    }
	    String s = (String) JOptionPane.showInputDialog (
		this, "Enter the image URL.",
		"Open Image URL", JOptionPane.INFORMATION_MESSAGE,
		null, null, chooserURL);
	    if (s == null) {
		return;	// User pressed cancel button in InputDialog.
	    }
	    chooserURL = s;
	    URL url = new URL (s);
	    Image image = Util.loadImage (url);
	    SegmentApp.installNewImage (
		SegmentApp.INVALID_IMAGE_ID, 
		SegmentApp.grayImage, 
		SegmentApp.invertImage,
		SegmentApp.flipImage,
		image,
		new File (url.getFile()).getName());

	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error opening image: " + e.getMessage(),
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    // Open image by ID.
    void openImageByID () {
	Util.assert (SegmentApp.mode == SegmentApp.CLIENT_MODE);
	try {
	    if (SegmentApp.unsavedChanges) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"There are unsaved changes which will be lost " +
			"if this operation continues! " +
			"Are you sure you want to acquire a new image before " +
			"first saving the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } else if (SegmentApp.mode == SegmentApp.CLIENT_MODE && SegmentApp.imageLoaded()) {
		int answer = JOptionPane.showConfirmDialog (
		    this, Util.lineBreak (
			"Are you sure you want to acquire a new image before " +
			"first submitting or aborting the current " +
			"segmentation? " +
			"", SegmentApp.infoWidth),
		    "Question", JOptionPane.YES_NO_OPTION);
		if (answer == JOptionPane.NO_OPTION) {
		    return;
		}
	    } 

	    String s = (String) JOptionPane.showInputDialog (
		this, "Enter the image ID.",
		"Open Image By ID", JOptionPane.INFORMATION_MESSAGE,
		null, null, "");
	    if (s == null) {
		// User pressed cancel button in InputDialog.
		return;
	    }
	    int imageID = Integer.decode(s).intValue();
	    if (imageID < 0) {
		throw new Exception ("Invalid image ID: " + imageID + ".");
	    }
	    Image image = SegmentApp.loadImageByID (imageID);
	    SegmentApp.installNewImage (imageID, 
					SegmentApp.grayImage, 
					SegmentApp.invertImage,
					SegmentApp.flipImage,
					image, null);

	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error opening image: " + e.getMessage(),
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    void submitSegmentation () {
	Util.assert (SegmentApp.mode == SegmentApp.CLIENT_MODE);
	try {
	    if (!SegmentApp.imageLoaded()) {
		JOptionPane.showMessageDialog (
		    this, "No segmentation to submit!", 
		    "Error", JOptionPane.ERROR_MESSAGE);
		return;
	    }
	    if (SegmentApp.unsavedChanges) {
		JOptionPane.showMessageDialog (
		    this, Util.lineBreak (
			"There are unsaved changes!  " +
			"You must save the segmentation before submitting. " +
			"", SegmentApp.infoWidth),
		    "Error", JOptionPane.ERROR_MESSAGE);
		return;
	    } 
	    Util.assert (SegmentApp.imageLoaded());
	    int answer = JOptionPane.showConfirmDialog (
		this, Util.lineBreak (
		    "Are you absolutely sure you want to submit this segmentation? " +
		    "\n" +
		    "You should be completely done with the segmentation, " +
		    "as you will not be able to edit it " +
		    "again after submitting!" +
		    "", SegmentApp.infoWidth),
		"Question", JOptionPane.YES_NO_OPTION);
	    if (answer != JOptionPane.YES_OPTION) {
		return;
	    }
	    SegmentApp.submitSegmentation ();
	    JOptionPane.showMessageDialog (
		this, "Segmentation successfully submitted.",
		"Message", JOptionPane.INFORMATION_MESSAGE);

	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error submitting segmentation: " + e.getMessage(),
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    void abortSegmentation () {
	Util.assert (SegmentApp.mode == SegmentApp.CLIENT_MODE);
	try {
	    if (!SegmentApp.imageLoaded()) {
		JOptionPane.showMessageDialog (
		    this, "No segmentation to abort!", 
		    "Error", JOptionPane.ERROR_MESSAGE);
		return;
	    }
	    Util.assert (SegmentApp.imageLoaded());
	    int answer = JOptionPane.showConfirmDialog (
		this, Util.lineBreak (
		    "Any unsaved changes will be lost after this operation! " +
		    "Are you absolutely sure you want to abort this segmentation? " + 
		    "", SegmentApp.infoWidth),
		"Question", JOptionPane.YES_NO_OPTION);
	    if (answer != JOptionPane.YES_OPTION) {
		return;
	    }
	    SegmentApp.abortSegmentation ();
	    JOptionPane.showMessageDialog (
		this, "Segmentation successfully aborted.", 
		"Message", JOptionPane.INFORMATION_MESSAGE);

	} catch (Exception e) {
	    JOptionPane.showMessageDialog (
		this, Util.lineBreak (
		    "Error aborting segmentation: " + e.getMessage(),
		    SegmentApp.infoWidth), 
		"Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    // Handle menu events.
    public void actionPerformed (ActionEvent e) {
	Object item = e.getSource ();
	if (item == editMenuUndo) {
	    SegmentApp.undo ();
	} else if (item == editMenuRedo) {
	    SegmentApp.redo ();
	} else if (item == fileMenuNewImage) {
	    getNewImage ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuChangeUserID) {
	    SegmentApp.changeUserID ();
	} else if (item == fileMenuOpenImageByID) {
	    openImageByID ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuOpenImageFile) {
	    openImageFile ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuOpenImageURL) {
	    openImageURL ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuOpenSegFile) {
	    openSegmentationFile ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuOpenSegURL) {
	    openSegmentationURL ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuSaveSeg) {
	    saveSegmentation ();
	} else if (item == fileMenuSaveAsSeg) {
	    saveAsSegmentation ();
	} else if (item == fileMenuSubmitSeg) {
	    submitSegmentation ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuOpenSubmittedSeg) {
	    openSubmittedSegmentation ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuOpenSubmittedSegByID) {
	    openSubmittedSegmentationByID ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuAbortSeg) {
	    abortSegmentation ();
	    updateImageDisplay ();
	    splitterWindow.updateImageDisplay ();
	} else if (item == fileMenuQuit) {
	    SegmentApp.quit ();
	} else if (item == viewMenuRefresh) {
	    // UI updated below.
	} else if (item == viewMenuGray) {
	    SegmentApp.grayImage = !SegmentApp.grayImage;
	    SegmentApp.segmentation.gray = SegmentApp.grayImage;
	    if (SegmentApp.imageLoaded()) {
		SegmentApp.computeDisplayImage ();
		updateImageDisplay ();
		splitterWindow.leftImagePanel.remaskImage ();
		splitterWindow.rightImagePanel.remaskImage ();
		splitterWindow.updateImageDisplay ();
	    }
	} else if (item == viewMenuFlip) {
	    SegmentApp.flipImage = !SegmentApp.flipImage;
	    SegmentApp.segmentation.flipflop = SegmentApp.flipImage;
	    if (SegmentApp.imageLoaded()) {
		SegmentApp.segmentation.doFlipFlop ();
		SegmentApp.computeDisplayImage ();
		updateImageDisplay ();
		splitterWindow.leftImagePanel.remaskImage ();
		splitterWindow.rightImagePanel.remaskImage ();
		splitterWindow.updateImageDisplay ();
	    }
	} else if (item == viewMenuInvert) {
	    SegmentApp.invertImage = !SegmentApp.invertImage;
	    SegmentApp.segmentation.invert = SegmentApp.invertImage;
	    if (SegmentApp.imageLoaded()) {
		SegmentApp.computeDisplayImage ();
		updateImageDisplay ();
		splitterWindow.leftImagePanel.remaskImage ();
		splitterWindow.rightImagePanel.remaskImage ();
		splitterWindow.updateImageDisplay ();
	    }
	} else if (item == viewMenuSegments) {
	    showSegments = !showSegments;
	    if (SegmentApp.imageLoaded()) {
		repaint ();
	    }
	} else if (item == viewMenuImage) {
	    showImage = !showImage;
	    if (SegmentApp.imageLoaded()) {
		repaint ();
	    }
	} else if (item == viewMenuFitWindow) {
	    fitWindow = viewMenuFitWindow.isSelected ();
	    if (fitWindow) {
		updateImageDisplay ();
	    }
	} else if (item == viewMenuMag1x) {
	    setMagnification (1);
	    fitWindow = false;
	    updateImageDisplay ();
	} else if (item == viewMenuMag2x) {
	    setMagnification (2);
	    fitWindow = false;
	    updateImageDisplay ();
	} else if (item == viewMenuMag3x) {
	    setMagnification (3);
	    fitWindow = false;
	    updateImageDisplay ();
	} else if (item == viewMenuMag4x) {
	    setMagnification (4);
	    fitWindow = false;
	    updateImageDisplay ();
	} else if (item == fitWindowCheckBox) {
	    fitWindow = fitWindowCheckBox.isSelected ();
	    if (fitWindow) {
		updateImageDisplay ();
	    }
	} else if (item == helpMenuFiles) {
	    helpFiles ();
	} else if (item == helpMenuEdit) {
	    helpEdit ();
	} else if (item == magTextField) {
	    // User entered a new magnification in the text area.
	    String text = magTextField.getText();
	    double newMag;
	    try {
		newMag = Double.parseDouble (text);
	    } catch (NumberFormatException exn) {
		newMag = magnification;
	    }
	    setMagnification (newMag);
	    fitWindow = false;
	    updateImageDisplay ();
	}
	updateUserInterface ();
    }
    
    void helpFiles () {
	final String message;
	if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
	    message =
	    	"The first thing you need to do is obtain a new image from the server. " +
		"Select the \"Get New Image\" command from the File menu.  Once you " +
		"have loaded an image, you will do one of three things: submit the " +
		"segmentation, abort the segmentation, or save the segmentation for " +
		"later. " +
		"\n" +
		"If you finish segmenting an image, then you should select the \"Submit " +
		"Segmentation\" command from the File menu.  This will send the " +
		"segmentation to the server.  You will not be able to edit your " +
		"segmentation again after submitting it, so be sure you are finished " +
		"before submitting it! " +
		"\n" +
		"If for some reason you don't want to segment an image that the server " +
		"gives you, then you can select the \"Cancel Segmentation\" option from " +
		"the File menu.  This will throw away the image you are working on. " +
		"Unless you have saved a local copy of your segmentation (see below), " +
		"it will be forever lost, so make sure this is what you want to do. " +
		"\n" +
		"If you need to save a segmentation mid-way through, then you can use " +
		"the \"Save Local Copy\" command in the File menu.  This will save the " +
		"segmentation in your home directory.  If you want to control where " +
		"segmentation files are saved, then use the \"Save Local Copy As...\" " +
		"command.  To resume this segmentation later, use the \"Open Local " +
		"Copy...\" command in the File menu. " +
		"\n" +
		"If you need to view or change a segmentation that you previously " +
		"submitted, then use the \"Open Submitted Segmentation...\" command " +
		"from the File menu.  The program will ask you to select an image ID " +
		"from the list of images for which you have submitted segmentations. " +
		"The images should be listed in the order in which you submitted them. " +
		"It is perfectly safe to change the segmentation and re-submit it. " +
		"If you know the image ID of the submitted segmentation you want to " +
		"edit, then you can use the \"Open Submitted Segmentation by ID...\" " +
		"command from the File menu. " +
		"\n";
	} else {
	    message =
		"The first thing you need to do is load an image.  From the file menu, " +
		"you can load image files or image URLs.  Once you have an image loaded " +
		"into the application, you can manipulate segmentations of that image. " +
		"\n" +
		"When you first load an image, the segmentation with all pixels in one " +
		"segment is created in memory.  If you want to create a new " +
		"segmentation from scratch for the image you just loaded, then you may " +
		"do so at this point.  Edit the segmentation, and save the segmentation " +
		"to a file. " +
		"\n" +
		"If you want to view or edit a pre-existing segmentation of the current " +
		"image, then you can load segmentations as files or URLs.  The " +
		"application will permit you to load a segmentation so long as the " +
		"width/height in the segmentation file match the width/height of the " +
		"current image. " +
		"\n" +
		"If you want to load a segmentation of an image that is not the current " +
		"image, then you must first load the other image before loading the " +
		"segmentation. " +
		"\n";
	}
	JOptionPane.showMessageDialog (
	    this, Util.lineBreak (message, SegmentApp.helpWidth),
	    "Help", JOptionPane.INFORMATION_MESSAGE);
    }

    void helpEdit () {
	final String message = 
	    "Once you have loaded an image, you may edit its segmentation.  Each " +
	    "segment in the image is displayed with a green boundary.  When you " +
	    "first load an image, all pixels belong to a single segment, which is " +
	    "why the image is drawn with a green boundary. " +
	    "\n" +
	    "In order to split a segment, click once on the segment.  This will " +
	    "open the segment splitter window, which will allow you to break the " +
	    "segment into two pieces.  See the help in that window for how to split " +
	    "a segment. " +
	    "\n" +
	    "Another important operation is moving pixels between two existing " +
	    "segments.  This is useful for cleaning up segment boundaries.  In " +
	    "order to do this, select one of the segments by clicking once in its " +
	    "interior.  Then click on the second segment while holding down the " +
	    "shift key.  This will select both segments.  The splitter window will " +
	    "now show these two segments, allowing you to move pixels between them. " +
	    "\n" +
	    "TIPS " +
	    "\n" +
	    "If you have trouble seeing the image below your segment boundaries, " +
	    "use the \"Show Segments\" command (Ctrl-W) from the View menu to turn " +
	    "the display of the segment boundaries on and off.  You can also toggle " +
	    "display of the image itself with \"Show Image\" in the View menu. " +
	    "\n";
	JOptionPane.showMessageDialog (
	    this, Util.lineBreak (message, SegmentApp.helpWidth),
	    "Help", JOptionPane.INFORMATION_MESSAGE);
    }

    // Preconditions: None.
    // Postconditions: Menu items are enbled/disabled in a manner
    // consistent with the data structures.
    void updateMenus () {
	fileMenuSaveSeg.setEnabled (SegmentApp.imageLoaded());
	fileMenuSaveAsSeg.setEnabled (SegmentApp.imageLoaded());
	if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
	    fileMenuAbortSeg.setEnabled (SegmentApp.imageLoaded());
	    fileMenuSubmitSeg.setEnabled (SegmentApp.imageLoaded());
	}
	if (SegmentApp.mode == SegmentApp.HERMIT_MODE) {
	    fileMenuOpenSegFile.setEnabled (SegmentApp.imageLoaded());
	    fileMenuOpenSegURL.setEnabled (SegmentApp.imageLoaded());
	}
	editMenuUndo.setEnabled (SegmentApp.undoActions.size() > 0);
	editMenuRedo.setEnabled (SegmentApp.redoActions.size() > 0);
	viewMenuFitWindow.setSelected (fitWindow);
	viewMenuGray.setSelected (SegmentApp.grayImage);
	viewMenuInvert.setSelected (SegmentApp.invertImage);
	viewMenuFlip.setSelected (SegmentApp.flipImage);
	fitWindowCheckBox.setSelected (fitWindow);
    }

    void updateLabels () {
	if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
	    userNameVal.setText (SegmentApp.userName + " (#" 
				 + SegmentApp.userID + ")");
	    totalImageCountVal.setText ("" + SegmentApp.totalImageCount);
	    sessionImageCountVal.setText ("" + SegmentApp.sessionImageCount);
	}
	if (SegmentApp.imageLoaded()) {
	    if (SegmentApp.mode == SegmentApp.CLIENT_MODE) {
		imageNameVal.setText (
		    "#" + SegmentApp.imageID 
		    + (SegmentApp.unsavedChanges ? " *unsaved*" : ""));
	    } else {
		imageNameVal.setText (
		    SegmentApp.imageFileName
		    + (SegmentApp.unsavedChanges ? " *unsaved*" : ""));
		if (SegmentApp.segFileName == null) {
		    segNameVal.setText ("<none>");
		} else {
		    segNameVal.setText (SegmentApp.segFileName);
		}
	    }
	} else {
	    imageNameVal.setText ("<none>");
	    segNameVal.setText ("<none>");
	}
	if (SegmentApp.imageLoaded()) {
	    segCountVal.setText ("" + SegmentApp.segmentation.nonEmptySegmentCount());
	} else {
	    segCountVal.setText ("0");
	}
    }

    // Always call this method to change the magnification.
    void setMagnification (double mag) {
	// Make sure magnification is in the valid range.
	mag = Math.max (mag, SegmentApp.minMagnification);
	mag = Math.min (mag, SegmentApp.maxMagnification);
	// Make sure magnification has <= 5 significant digits
	// for display purposes.
	mag = (double) ((int) (mag * 10000.)) / 10000.;
	magnification = mag;
    }

    // Return the magnification that causes the image to fit perfectly
    // in the viewport.
    double magFromViewportSize () {
	Util.assert (SegmentApp.imageLoaded());
	double mag;
	Dimension dim = scrollPane.getViewport().getSize();
	if (dim.width < SegmentApp.imWidth * SegmentApp.minMagnification + 2 * border
	    || dim.height < SegmentApp.imHeight * SegmentApp.minMagnification + 2 * border) {
	    mag = SegmentApp.minMagnification;
	} else {
	    double imageAspectRatio = (double) SegmentApp.imWidth / SegmentApp.imHeight;
	    double panelAspectRatio = (double) (dim.width - 2 * border) 
				      / (dim.height - 2 * border);
	    if (imageAspectRatio > panelAspectRatio) {
		mag = (double) (dim.width - 2 * border) / SegmentApp.imWidth;
	    } else {
		mag = (double) (dim.height - 2 * border) / SegmentApp.imHeight;
	    }
	}
	return mag;
    }

    // Preconditions: None.
    // Postconditions: User interface is consistent with data structures.
    void updateUserInterface () {
	magTextField.setText (String.valueOf (magnification));
	updateMenus ();
	updateLabels ();
	validate ();
	repaint ();
    }
    void updateImageDisplay () {
	if (SegmentApp.imageLoaded() && fitWindow) {
	    setMagnification (magFromViewportSize ());
	}
	segPanel.invalidate ();
	updateUserInterface ();
    }

    // Class to display displayImage in a panel.
    // Preferred size is the size of the image.
    class SimpleImagePanel 
	extends JPanel
    {
	public SimpleImagePanel () {
	    super ();
	}

	public Dimension getPreferredSize () {
	    if (SegmentApp.imageLoaded()) {
		return new Dimension (
		    (int) Math.round ((double)SegmentApp.imWidth * magnification + 2 * border), 
		    (int) Math.round ((double)SegmentApp.imHeight * magnification + 2 * border));
	    } else {
		return super.getPreferredSize ();
	    }
	}

	public void paint (Graphics g) {
	    super.paint (g);
	    if (SegmentApp.imageLoaded() && showImage) {
		int w = (int) Math.round ((double) SegmentApp.imWidth * magnification);
		int h = (int) Math.round ((double) SegmentApp.imHeight * magnification);
		boolean loaded = g.drawImage (
		    SegmentApp.displayImage, border, border, w, h, null);
		Util.assert (loaded);
	    }
	}
    }

    // Class to display displayImage and its segmentation in a panel.
    class SegmentationPanel 
	extends SimpleImagePanel
	implements MouseListener
    {
	public SegmentationPanel () {
	    super ();
	    addMouseListener (this);
	}

	// Handle mouse events.
	public void mouseReleased (MouseEvent e) {}
	public void mouseEntered (MouseEvent e) {}
	public void mouseExited (MouseEvent e) {}
	public void mouseClicked (MouseEvent e) {}
	public void mousePressed (MouseEvent e) {
	    if (!SegmentApp.imageLoaded()) {
		return;
	    }
	    Point p = e.getPoint();
	    int x = (int) Math.round ((double) (p.x - border) / magnification);
	    int y = (int) Math.round ((double) (p.y - border) / magnification);
	    if (x < 0 || x >= SegmentApp.imWidth ||
		y < 0 || y >= SegmentApp.imHeight) {
		// Any click outside of image deselects all segments.
		SegmentApp.selection.selectAll (false);
	    } else {
		int seg = SegmentApp.segmentation.get (x, y);
		if ((e.getModifiers() & InputEvent.SHIFT_MASK) != 0) {
		    // Shift-click toggles selection of target segment.
		    SegmentApp.selection.toggleSelection (seg);
		    SegmentApp.selection.deselectEmptySegments ();
		} else {
		    // Any other click inside image selects single segment.
		    SegmentApp.selection.selectAll (false);
		    SegmentApp.selection.toggleSelection (seg);
		}
	    }
	    int count = SegmentApp.selection.countSelected ();
	    switch (count) {
		case 0:
		    splitterWindow.clearSegments ();
		    break;
		case 1:
		    splitSegment ();
		    break;
		case 2:
		    editSegments ();
		    break;
		default:
		    splitterWindow.clearSegments ();
	    }
	    splitterWindow.updateImageDisplay ();
	    updateImageDisplay ();
	    splitterWindow.setState (Frame.NORMAL);
	    //splitterWindow.show ();
	    //splitterWindow.toFront ();
	}

	public void paint (Graphics g) {
	    super.paint (g);
	    if (SegmentApp.imageLoaded() && showSegments) {
		// It's very important to clip this computation, since the
		// computation is a bit slow for smooth scrolling.
		Rectangle clip = g.getClipBounds();
		int xmin = Math.max (0, (int) Math.round ((double) (clip.x - border) / magnification - 1));
		int xmax = Math.min ((int) Math.round ((double) (clip.x + clip.width - border) / magnification) + 1, SegmentApp.imWidth);
		int ymin = Math.max (0, (int) Math.round ((double) (clip.y - border) / magnification) - 1);
		int ymax = Math.min ((int) Math.round ((double) (clip.y + clip.height - border) / magnification) + 1, SegmentApp.imHeight);
		for (int y = ymin; y < ymax; y++) {
		    for (int x = xmin; x < xmax; x++) {
			int seg = SegmentApp.segmentation.get (x, y);
			int segN = (y == 0) ? -1 : SegmentApp.segmentation.get (x, y - 1);
			int segS = (y == SegmentApp.imHeight - 1) ? -1 : SegmentApp.segmentation.get (x, y + 1);
			int segW = (x == 0) ? -1 : SegmentApp.segmentation.get (x - 1, y);
			int segE = (x == SegmentApp.imWidth - 1) ? -1 : SegmentApp.segmentation.get (x + 1, y);
			if (seg == segN && seg == segS && seg == segW && seg == segE) {
			    // This test boosts performace quite a bit.
			    continue;
			}
			int x1 = (int) Math.round ((double) x * magnification + border);
			int y1 = (int) Math.round ((double) y * magnification + border);
			int x2 = (int) Math.round ((double) (x + 1) * magnification + border);
			int y2 = (int) Math.round ((double) (y + 1) * magnification + border);
			x2 = Math.max (x1, x2 - 1);
			y2 = Math.max (y1, y2 - 1);
			g.setColor (SegmentApp.selection.selected (seg) 
				    ? Color.red : Color.green);
			if (seg != segN) { g.drawLine (x1, y1, x2, y1); }
			if (seg != segS) { g.drawLine (x1, y2, x2, y2); }
			if (seg != segW) { g.drawLine (x1, y1, x1, y2); }
			if (seg != segE) { g.drawLine (x2, y1, x2, y2); }
		    }
		}
	    }
	}
    }
}
