
// Author:  David Martin  <dmartin@eecs.berkeley.edu>

// A namespace class for routines that transform ARGB pixels.
final class RGBFilter
{
    static private int clamp (int v) {
	if (v < 0) { v = 0; }
	if (v > 255) { v = 255; }
	return v;
    }

    static public int lighten (int rgb) {
	// lightness=1 --> no brigtness change
	// lightness=0 --> white
	final double lightness = 0.2;

	// redness=1 --> no hue change
	// redness=0 --> red
	final double redness = 0.65;

	int a = (rgb >> 24) & 0xff;
	int r = (rgb >> 16) & 0xff;
	int g = (rgb >>  8) & 0xff;
	int b = (rgb >>  0) & 0xff;

	r = (int) (255. - (255 - r) * lightness);
	g = (int) (255. - (255 - g) * lightness);
	b = (int) (255. - (255 - b) * lightness);

	g = (int) (g / 255. * redness * 255.);
	b = (int) (b / 255. * redness * 255.);

	r = clamp (r);
	g = clamp (g);
	b = clamp (b);

	return a << 24 | r << 16 | g << 8 | b;
    }

    static public int toGray (int rgb) {
	int a = (rgb >> 24) & 0xff;
	int r = (rgb >> 16) & 0xff;
	int g = (rgb >>  8) & 0xff;
	int b = (rgb >>  0) & 0xff;

	//r = g = b = (299 * r + 587 * g + 114 * b) / 1000;
	r = g = b = (306 * r + 601 * g + 117 * b) >> 10;

	return a << 24 | r << 16 | g << 8 | b;
    }

    static public int invert (int rgb) {
	int a = (rgb >> 24) & 0xff;
	int r = (rgb >> 16) & 0xff;
	int g = (rgb >>  8) & 0xff;
	int b = (rgb >>  0) & 0xff;

	r = 255 - r;
	g = 255 - g;
	b = 255 - b;

	return a << 24 | r << 16 | g << 8 | b;
    }

}

