
// Author:  David Martin  <dmartin@eecs.berkeley.edu>

import java.util.*;
import java.io.*;
import java.net.*;
import gnu.regexp.*;

public class SegServer
{
    String host;
    int port;
    Socket socket;
    DataOutputStream outputStream;
    BufferedReader reader;

    public SegServer (String host, int port) {
	this.host = host;
	this.port = port;
    }

    public void connect () throws Exception {
	InetAddress iaddr = InetAddress.getByName (host);
	socket = new Socket (iaddr, port);
	InputStream in = socket.getInputStream ();
	OutputStream out = socket.getOutputStream ();
	outputStream = new DataOutputStream (out);
	reader = new BufferedReader (new InputStreamReader (in));
    }

    public void disconnect () {
	try {
	    if (socket != null) { socket.close (); } 
	} catch (Exception e) {
	}
    }

    public void sendLine (String s) throws Exception {
	outputStream.writeBytes (s);
	outputStream.writeBytes ("\n");
    }

    public void sendDone () throws Exception {
	sendLine ("eof");
	//socket.shutdownOutput ();
    }

    // Parse server reply into a hash object.
    public Map getReply () throws Exception {
	Map map = new HashMap ();
	RE re;
	REMatch rem;
	String line;

	// First line of reply should be 'ok' or 'error'.
	line = reader.readLine ();
	if (line.startsWith ("error")) {
	    throw new Exception ("Server reply error.");
	} 
	if (!line.startsWith ("ok")) {
	    throw new Exception ("Invalid server reply: " + line);
	}

	// Reply is a sequence of key/value pairs.
	while (true) {
	    line = reader.readLine ();
	    if (line == null) { 
		throw new Exception ("Premature EOF in server reply.");
	    }

	    // Clean up line.
	    line = new RE ("\\s+").substituteAll (line, " ");	// Compress whitespace.
	    line = new RE ("^ ").substitute (line, "");		// Remove leading whitespace.
	    line = new RE (" $").substitute (line, "");		// Remove trailing whitespace.

	    // 'eof' marks end of reply.
	    re = new RE ("^eof");
	    if (re.isMatch (line)) {
		break;
	    }

	    // Begin/end blocks provide list values.
	    re = new RE ("^begin (\\S+)$");
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		String key = rem.toString (1);
		List data = new ArrayList ();
		while (true) {
		    line = reader.readLine ();
		    if (line == null) { break; }
		    if (line.startsWith ("end")) { break; }
		    re = new RE ("^-(.*)$");
		    if (re.isMatch (line)) {
			rem = re.getMatch (line);
			String v = rem.toString (1);
			data.add (v);
		    } else {
			throw new Exception ("Parse error in 'begin " + 
					     key + "'...end block: " + line);
		    }
		}
		map.put (key, data);
		continue;
	    }

	    // Plain key/value pair.
	    re = new RE ("^(\\S+) (.*)$");
	    if (re.isMatch (line)) {
		rem = re.getMatch (line);
		String key = rem.toString (1);
		String value = rem.toString (2);
		map.put (key, value);
		continue;
	    }
	    
	    throw new Exception ("Parse error: " + line);
	}
	return map;
    }

}

